#include <cstddef>
#include <cstdio>
#include <fstream>
#include <windef.h>
#include <errhandlingapi.h>
#include <libloaderapi.h>

using namespace std;

template <typename T>
class TSArray {
public:
	size_t capacity = 0;
	size_t size = 0;
	T* items = nullptr;
};

using CDKey = char[32];
using CDKeys = TSArray<CDKey>;

typedef BOOL (__fastcall* TReadCredentials)(CDKeys* keys, char *owner_name, BOOL *b);
static TReadCredentials ReadCredentials;

static const char* LOG_PATH = "./cd_key_dumper.log";
static ofstream log;

static bool initialize_log() {
	log.open(LOG_PATH);
	if (!log) {
		fprintf(
			stderr,
			"[CDKeyDumper] Failed to create log file at «%s».",
			LOG_PATH
		);
		return false;
	}
	else {
		return true;
	}
}

static bool initialize_offsets() {
	HMODULE game_base = GetModuleHandle("game.dll");
	if (!game_base) {
		fprintf(
			stderr,
			"[CDKeyDumper] Failed to get the «game.dll» module handle. Error = %d",
			GetLastError()
		);
		return false;
	}

	ReadCredentials = (TReadCredentials)(((char*)game_base) + 0x553CC0);

	return true;
}

static void print_cd_key(ofstream& stream, CDKey& key) {
	char buffer[sizeof(CDKey) + 1];
	memcpy(buffer, key, sizeof(CDKey));
	buffer[sizeof(CDKey)] = '\0';
	stream << buffer;
}

static bool dump_cd_keys() {
	CDKeys keys;
	char owner_name[64];
	BOOL b;

	if (ReadCredentials(&keys, owner_name, &b)) {
		log << "owner name: «" << owner_name << "»\n";
		log << "unknown boolean: " << b << "\n";

		for (int i = 0; i < keys.size; i++) {
			log << "CD-Key " << i << ": «";
			print_cd_key(log, keys.items[i]);
			log << "»\n";
		}

		log.flush();

		return true;
	}
	else {
		log << "Failed to get credentials." << endl;
		return false;
	}
}

extern "C"
BOOL APIENTRY DllMain(HMODULE hModule, DWORD fdwReason, LPVOID lpvReserved) {
	switch (fdwReason) {
		case DLL_PROCESS_ATTACH: {
			return initialize_log()
				&& initialize_offsets()
				&& dump_cd_keys();
				;
		}
	}

	return true;
}